from plone.app.testing import PloneSandboxLayer, PLONE_FIXTURE, \
    IntegrationTesting
from plone.testing import z2
from zope.configuration import xmlconfig
from plone.app.testing import applyProfile


class CollectiveCeleryLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        import collective.celery
        xmlconfig.file(
            'configure.zcml',
            collective.celery,
            context=configurationContext
        )
        z2.installProduct(app, 'collective.celery')

    def tearDownZope(self, app):
        z2.uninstallProduct(app, 'collective.celery')

    def setUpPloneSite(self, portal):
        # Install into Plone site using portal_setup
        applyProfile(portal, 'plone.app.contenttypes:default')


COLLECTIVE_CELERY_FIXTURE = CollectiveCeleryLayer()

COLLECTIVE_CELERY_INTEGRATION_TESTING = IntegrationTesting(
    bases=(COLLECTIVE_CELERY_FIXTURE,),
    name="CollectiveCeleryLayer:Integration"
)



# a lot of this pulled out of pyramid_celery
from importlib import import_module
import os
import sys

from App.config import getConfiguration
from celery.bin.celery import CeleryCommand
from celery.utils.log import get_task_logger
from collective.celery.utils import getCelery
from pkg_resources import iter_entry_points


logger = get_task_logger(__name__)


class CommandMixin(object):
    preload_options = ()

    def setup_app_from_commandline(self, argv):
        self.app = getCelery()
        return argv


class Worker(CommandMixin, CeleryCommand):
    pass


def main(argv=sys.argv):
    if len(sys.argv) < 3:
        raise Exception("must specify a zope config file and a celery command")
    argv = argv

    # find the index of the conf file in the args
    conf_index = 2
    for idx, arg in enumerate(sys.argv):
        if '.conf' in arg:
            conf_index = idx
            break
    filepath = sys.argv[conf_index]
    os.environ['ZOPE_CONFIG'] = filepath
    sys.argv = ['']
    from Zope2.Startup.run import configure
    startup = configure(os.environ['ZOPE_CONFIG'])

    # Fix for setuptools generated scripts, so that it will
    # work with multiprocessing fork emulation.
    # (see multiprocessing.forking.get_preparation_data())
    if __name__ != "__main__":
        sys.modules["__main__"] = sys.modules[__name__]

    # load entry point tasks up
    tasks = []
    for entry_point in iter_entry_points(group='celery_tasks', name=None):
        try:
            tasks.append((entry_point.name, entry_point.load()))
        except ImportError:
            logger.warn('error importing tasks: ' + entry_point.name)
            raise
    tasks = dict(tasks)
    for name, task_list in tasks.items():
        logger.warn('importing tasks: ' + name)
        extra_config = getattr(task_list, 'extra_config', None)
        if extra_config is not None:
            logger.warn('Found additional Zope config.')
            extra_config(startup)

    # load env tasks up
    tasks = getConfiguration().environment.get('CELERY_TASKS')
    if tasks:
        for task_list in tasks.split():
            try:
                logger.warn('importing tasks: ' + tasks)
                module = import_module(tasks)
                extra_config = getattr(module, 'extra_config', None)
                if extra_config is not None:
                    logger.warn('Found additional Zope config.')
                    extra_config(startup)
            except ImportError:
                logger.warn('error importing tasks: ' + tasks)
                raise
    argv.remove(filepath)
    # restore argv
    sys.argv = argv
    Worker(app=getCelery()).execute_from_commandline()



import OFS
from plone import api

from .base import BaseTestCase
from ..utils import _serialize_arg, _deserialize_arg, _bool, getCeleryOptions, getApp
import os


class TestHelpers(BaseTestCase):

    def setUp(self):
        super(TestHelpers, self).setUp()
        self.login_as_portal_owner()
        self.doc = api.content.create(
            self.portal,
            'Document',
            'doc'
        )

    def test__serialize_arg(self):
        result = _serialize_arg(self.doc)
        expected = 'object:///plone/doc'
        self.assertEqual(
            result,
            expected,
            'Serializer not returning expected output'
        )

    def test__deserialize_arg(self):
        result = _deserialize_arg(self.portal, 'object:///plone/doc')
        self.assertEqual(
            result,
            self.doc,
            'Did not get the expected object when deserializing'
        )

    def test__bool(self):
        self.assertTrue(_bool('1'))
        self.assertTrue(_bool('true'))
        self.assertTrue(_bool('yes'))
        self.assertFalse(_bool('0'))
        self.assertFalse(_bool('false'))
        self.assertFalse(_bool('no'))

    def test_getCeleryOptions(self):
        os.environ['CELERY_TEST'] = 'True'
        result = getCeleryOptions()
        self.assertIn('CELERY_TEST', result)

    def test_getApp(self):
        app = getApp()
        self.assertIsInstance(app, OFS.Application.Application)


from App.config import getConfiguration
from celery import current_app as celery
from celery.app import defaults
from OFS.interfaces import IItem

import os
import sys
import threading
import Zope2


try:
    from celery import registry
except ImportError:
    # Celery >= 3.1
    from celery import current_app as registry


_local = threading.local()


def _bool(term, table={"false": False, "no": False, "0": False,
                       "true": True, "yes": True, "1": True}):
    try:
        return table[term.lower()]
    except KeyError:
        raise TypeError("Can't coerce %r to type bool" % (term, ))


_types = {
    'any': (object, None),
    'bool': (bool, _bool),
    'dict': (dict, eval),
    'float': (float, float),
    'int': (int, int),
    'list': (list, eval),
    'tuple': (tuple, eval),
    'string': (str, str),
    'str': (str, str),
}

_options = dict(
    (key, _types[opt.type])
    for key, opt in defaults.flatten(defaults.NAMESPACES)
    if opt.type in _types
)

_object_marker = 'object://'


def getCeleryOptions():
    zconfig = getConfiguration()
    if hasattr(zconfig, 'environment'):
        environ = zconfig.environment.items()
    else:
        # sort of for testing...
        environ = os.environ.items()

    config = {}
    for key, value in environ:
        opt_type = _options.get(key)
        if opt_type:
            if opt_type[0] == str:
                value = value.replace('"', '')
            elif opt_type[0] is object:
                try:
                    value = eval(value)
                except:
                    pass  # any can be anything; even a string
            elif not isinstance(value, opt_type[0]):
                value = opt_type[1](value)
        config[key] = value
    return config


def _getCelery():
    celery.add_defaults(getCeleryOptions())
    # delete cached property in order to get them reloaded from the new conf
    del(celery.backend)
    for name, task in registry.tasks.items():
        # ensure that every already registed tasks doens use an unconfigured
        # backend.
        task.backend = celery.backend
    return celery


def getCelery():
    if not hasattr(_local, 'celery'):
        _local.celery = _getCelery()
    return _local.celery


def setApp(app):
    _local.app = app


def getApp(*args, **kwargs):
    try:
        return _local.app
    except AttributeError:
        pass
    if Zope2.bobo_application is None:
        orig_argv = sys.argv
        sys.argv = ['']
        res = Zope2.app(*args, **kwargs)
        sys.argv = orig_argv
        return res
    # should set bobo_application
    # man, freaking zope2 is weird
    return Zope2.bobo_application(*args, **kwargs)


def _serialize_arg(val):
    if IItem.providedBy(val):
        val = '%s%s' % (
            _object_marker,
            '/'.join(val.getPhysicalPath()))
    return val


def _deserialize_arg(app, val):
    if isinstance(val, basestring):
        if val.startswith(_object_marker):
            val = val[len(_object_marker):]
            val = app.unrestrictedTraverse(val)
    return val



from plone import api
from ..base_task import EagerResult, AfterCommitTask
from .base import BaseTestCase

__author__ = 'benc'


class TestEagerResult(BaseTestCase):

    def test_ready(self):
        result = EagerResult('1234', 'Done', 'SUCCESS')
        self.assertTrue(result.ready())


class TestAfterCommitTask(BaseTestCase):

    def setUp(self):
        super(TestAfterCommitTask, self).setUp()
        self.login_as_portal_owner()
        self.doc1 = api.content.create(
            self.portal,
            'Document',
            'doc1'
        )
        self.doc2 = api.content.create(
            self.portal,
            'Document',
            'doc2'
        )
        self.task = AfterCommitTask()

    def test_serialize_args(self):
        args = [self.doc1]
        kwargs = {'doc': self.doc2}
        result = self.task.serialize_args(args, kwargs)
        expected = (['object:///plone/doc1'], {'doc': 'object:///plone/doc2'})
        self.assertEqual(
            result,
            expected,
            'AfterCommitTask.serialize_args return value not expected'
        )

    def test_apply_async(self):
        pass


# -*- encoding: utf-8 -*-
# This is all pulled out of David Glick's gist on github
# https://gist.githubusercontent.com/davisagli/5824662/raw/de6ac44c1992ead62d7d98be96ad1b55ed4884af/gistfile1.py
from base_task import AfterCommitTask
from celery import current_app
from celery.signals import after_task_publish
from collective.celery.functionrunner import AdminFunctionRunner
from collective.celery.functionrunner import AuthorizedFunctionRunner
from collective.celery.utils import getCelery


TESTING = False


def initialize(context):
    pass


class _task(object):
    """Decorator of celery tasks that should be run in a Zope context.

    The decorator function takes a path as a first argument,
    and will take care of traversing to it and passing it
    (presumably a portal) as the first argument to the decorated function.

    Also takes care of initializing the Zope environment,
    running the task within a transaction, and retrying on
    ZODB conflict errors.
    """

    def __call__(self, **task_kw):
        def decorator(func):
            def new_func(*args, **kw):
                runner = AuthorizedFunctionRunner(func, new_func, args, kw, task_kw)  # noqa
                return runner()
            new_func.__name__ = func.__name__
            return getCelery().task(base=AfterCommitTask, **task_kw)(new_func)
        return decorator

    def as_admin(self, **task_kw):
        def decorator(func):
            def new_func(*args, **kw):
                runner = AdminFunctionRunner(func, new_func, args, kw, task_kw)
                return runner()
            new_func.__name__ = func.__name__
            return getCelery().task(base=AfterCommitTask, **task_kw)(new_func)
        return decorator

task = _task()
task.__doc__ = """This decorator "wraps" the celery task decorator
:py:meth:`celery.app.base.Celery.task`.

It can be used the same way::

    @task()
    def mytask():
        pass

or through the additional ``as_admin()`` method::

    @task.as_admin()
    def mytask():
        pass

Which will execute the task in an unrestricted environment.
"""


@after_task_publish.connect
def update_sent_state(sender=None, body=None, **kwargs):
    """so we can know if a task was scheduled"""
    task = current_app.tasks.get(sender)
    task.update_state(task_id=kwargs['headers']['id'], state="SENT")



from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from celery.exceptions import Retry
from celery.utils.log import get_task_logger
from collective.celery.base_task import AfterCommitTask
from collective.celery.utils import _deserialize_arg
from collective.celery.utils import getApp
from collective.celery.utils import getCelery
from plone import api
from Testing.makerequest import makerequest
from ZODB.POSException import ConflictError
from zope.app.publication.interfaces import BeforeTraverseEvent
from zope.component.hooks import setSite
from zope.event import notify
from zope.globalrequest import clearRequest
from zope.globalrequest import setRequest

import traceback
import transaction


logger = get_task_logger(__name__)


class FunctionRunner(object):

    base_task = AfterCommitTask
    app = None
    eager = False

    def __init__(self, func, new_func, orig_args, orig_kw, task_kw):
        self.orig_args = orig_args
        self.orig_kw = orig_kw
        self.func = func
        self.new_func = new_func
        self.userid = None
        self.site = None
        self.app = None
        self.task_kw = task_kw

    def deserialize_args(self):
        args = []
        kw = {}
        for arg in self.orig_args:
            args.append(_deserialize_arg(self.app, arg))
        for key, value in self.orig_kw.items():
            kw[key] = _deserialize_arg(self.app, value)

        return args, kw

    def authorize(self):
        pass

    def _run(self):
        self.userid = self.orig_kw.pop('authorized_userid')
        site_path = self.orig_kw.pop('site_path')
        if self.app is None:
            self.site = api.portal.get()
        else:
            self.site = self.app.unrestrictedTraverse(site_path)
        self.authorize()
        args, kw = self.deserialize_args()  # noqa
        # run the task
        return self.func(*args, **kw)

    def __call__(self):
        celery = getCelery()
        if celery.conf.CELERY_ALWAYS_EAGER:
            self.eager = True
            # dive out of setup, this is not run in a celery task runner
            self.app = getApp()
            return self._run()

        self.app = makerequest(getApp())
        self.app.REQUEST['PARENTS'] = [self.app]
        setRequest(self.app.REQUEST)

        transaction.begin()
        try:
            try:
                result = self._run()
                # commit transaction
                transaction.commit()
                return result
            except ConflictError, e:
                # On ZODB conflicts, retry using celery's mechanism
                transaction.abort()
                raise Retry(exc=e)
            except:
                logger.warn('Error running task: %s' % traceback.format_exc())
                transaction.abort()
                raise
        finally:
            noSecurityManager()
            setSite(None)
            self.app._p_jar.close()
            clearRequest()


class AuthorizedFunctionRunner(FunctionRunner):

    def authorize(self):
        if self.eager:
            # ignore, run as current user
            return

        notify(BeforeTraverseEvent(self.site, self.site.REQUEST))
        setSite(self.site)

        # set up user
        # TODO: using plone.api.get_user().getUser()
        # somehow makes the test fail, probably because the whole setRoles
        # and login() don't do everything.
        user = api.user.get(userid=self.userid).getUser()
        newSecurityManager(None, user)


class AdminFunctionRunner(AuthorizedFunctionRunner):

    def authorize(self):
        if self.eager:
            # ignore, run as current user
            return

        notify(BeforeTraverseEvent(self.site, self.site.REQUEST))
        setSite(self.site)

        # set up admin user
        # XXX need to search for an admin like user otherwise?
        user = api.user.get(userid='admin').getUser()
        if user:
            newSecurityManager(None, user)



from celery.app.task import TaskType
from plone.app.testing import SITE_OWNER_NAME
from zope.component import eventtesting

from .base import BaseTestCase
from ..functionrunner import FunctionRunner, AuthorizedFunctionRunner, AdminFunctionRunner


class BaseFunctionRunnerTestCase(BaseTestCase):

    def setUp(self):
        super(BaseFunctionRunnerTestCase, self).setUp()
        eventtesting.setUp()

        def dummy_func(*args, **kwargs):
            return 'Bar'

        self.dummy_func = dummy_func
        self.args = ['foo', 'bar']
        self.kwargs = {
            'authorized_userid': SITE_OWNER_NAME,
            'site_path': '/plone',
        }
        self.task_kwargs={'baz': 1}
        self.frunner = FunctionRunner(
            func=self.dummy_func,
            new_func=self.dummy_func,
            orig_args=self.args,
            orig_kw=self.kwargs,
            task_kw=self.task_kwargs,
        )


class TestFunctionRunner(BaseFunctionRunnerTestCase):

    def test__init(self):
        self.assertIsInstance(self.frunner.base_task, TaskType)
        self.assertIs(self.frunner.func, self.dummy_func)
        self.assertIs(self.frunner.new_func, self.dummy_func)
        self.assertEqual(self.frunner.orig_args, self.args)
        self.assertEqual(self.frunner.orig_kw, self.kwargs)
        self.assertEqual(self.frunner.task_kw, self.task_kwargs)
        self.assertIsNone(self.frunner.userid)
        self.assertIsNone(self.frunner.site)
        self.assertIsNone(self.frunner.app)

    def test__call(self):
        self.assertIs(self.frunner(), self.dummy_func())

    def test_deserialize_args(self):
        self.frunner()
        result = self.frunner.deserialize_args()
        self.assertEqual(result[0], self.args)
        self.assertEqual(result[1], self.kwargs)

    def test_authorize(self):
        self.assertIsNone(self.frunner.authorize())


class TestAuthorizedFunctionRunner(BaseFunctionRunnerTestCase):

    def setUp(self):
        super(TestAuthorizedFunctionRunner, self).setUp()
        self.frunner = AuthorizedFunctionRunner(
            func=self.dummy_func,
            new_func=self.dummy_func,
            orig_args=self.args,
            orig_kw=self.kwargs,
            task_kw=self.task_kwargs,
        )

    def test_authorize(self):
        # TODO: This is code does not work due to the fact we're starting a
        # Zope instance within a zope instance
        # self.frunner()
        # eventtesting.clearEvents()
        # self.frunner.authorize()
        # events = eventtesting.getEvents()
        # self.assertGreater(len(events), 0)
        pass


class TestAdminFunctionRunner(BaseFunctionRunnerTestCase):

    def setUp(self):
        super(TestAdminFunctionRunner, self).setUp()
        self.frunner = AdminFunctionRunner(
            func=self.dummy_func,
            new_func=self.dummy_func,
            orig_args=self.args,
            orig_kw=self.kwargs,
            task_kw=self.task_kwargs,
        )

    def test_authorize(self):
        # TODO: This is code does not work due to the fact we're starting a
        # Zope instance within a zope instance
        # self.frunner()
        # eventtesting.clearEvents()
        # self.frunner.authorize()
        # events = eventtesting.getEvents()
        # self.assertGreater(len(events), 0)
        pass


try:
    import unittest2 as unittest
except ImportError:
    import unittest
from plone.testing import z2
from plone.app.testing.interfaces import SITE_OWNER_NAME

from collective.celery import getCelery
from collective.celery.testing import COLLECTIVE_CELERY_INTEGRATION_TESTING
from collective.celery.utils import _getCelery
from collective.celery.utils import setApp


class BaseTestCase(unittest.TestCase):

    layer = COLLECTIVE_CELERY_INTEGRATION_TESTING

    def setUp(self):
        getCelery().conf['CELERY_ALWAYS_EAGER'] = 'True'
        # use in-memory sqlite
        getCelery().conf['BROKER_URL'] = "sqla+sqlite://"
        getCelery().conf['CELERY_RESULT_BACKEND'] = "db+sqlite://"
        # refresh cached properties
        _getCelery()
        self.app = self.layer['app']
        self.portal = self.layer['portal']

        setApp(self.app)

    def login_as_portal_owner(self):
        """
        helper method to login as site admin
        """
        z2.login(self.app['acl_users'], SITE_OWNER_NAME)



from setuptools import setup, find_packages
import os

version = '1.1.0'

setup(name='collective.celery',
      version=version,
      description="Celery for Plone",
      long_description="%s\n%s" % (
          open("README.rst").read(),
          open(os.path.join("docs", "CHANGES.rst")).read()
      ),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
          "Programming Language :: Python",
          "Framework :: Plone :: 4.0",
          "Framework :: Plone :: 4.1",
          "Framework :: Plone :: 4.2",
          "Framework :: Plone :: 4.3",
          "Framework :: Plone :: 5.0",
      ],
      keywords='celery async plone',
      author='Nathan Van Gheem',
      author_email='vangheem@gmail.com',
      url='https://github.com/collective/collective.celery',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['collective'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'celery',
          'plone.api'
      ],
      extras_require={
          'test': [
              'plone.app.testing',
              'SQLAlchemy'
          ]
      },
      entry_points="""
      [z3c.autoinclude.plugin]
      target = plone

      [console_scripts]
      pcelery = collective.celery.scripts.ccelery:main
      """,
      )



from celery import result
from celery import states
from celery import Task
from collective.celery.utils import _serialize_arg
from collective.celery.utils import getCelery
from kombu.utils import uuid
from plone import api
from transaction.interfaces import ISynchronizer
from zope.interface import implementer

import transaction


class EagerResult(result.EagerResult):

    def ready(self):
        return self._state in states.READY_STATES


@implementer(ISynchronizer)
class CelerySynchronizer(object):
    """Handles communication with celery at transaction boundaries.
    We previously used after-commit hooks, but the transaction package
    swallows exceptions in commit hooks.
    """

    def beforeCompletion(self, txn):
        pass

    def afterCompletion(self, txn):
        """Called after commit or abort
        """
        # Skip if running tests
        import collective.celery
        if collective.celery.TESTING:
            return False
        if txn.status == transaction._transaction.Status.COMMITTED:
            tasks = getattr(txn, '_celery_tasks', [])
            executed = []
            for args, kw, task, task_id, options in tasks:
                if (args, kw, options) in executed:
                    # make sure task was not sent multiple times
                    # by ignoring tasks with exact same args.
                    continue
                executed.append((args, kw, options))
                super(AfterCommitTask, task).apply_async(
                    args=args,
                    kwargs=kw,
                    task_id=task_id,
                    **options
                )

    def newTransaction(self, txn):
        pass

celery_synch = CelerySynchronizer()


def queue_task_after_commit(args, kw, task, task_id, options):
    transaction.manager.registerSynch(celery_synch)

    txn = transaction.get()
    if not hasattr(txn, '_celery_tasks'):
        txn._celery_tasks = []
    txn._celery_tasks.append((args, kw, task, task_id, options))


class AfterCommitTask(Task):
    """Base for tasks that queue themselves after commit.

    This is intended for tasks scheduled from inside Zope.
    """
    abstract = True

    def serialize_args(self, orig_args, orig_kw):
        args = []
        kw = {}
        for arg in orig_args:
            args.append(_serialize_arg(arg))
        for key, value in orig_kw.items():
            kw[key] = _serialize_arg(value)
        return args, kw

    # Override apply_async to register an after-commit hook
    # instead of queueing the task right away and to
    # set object paths instead of objects
    def apply_async(self, args, kwargs, **options):
        args, kw = self.serialize_args(args, kwargs)
        kw['site_path'] = '/'.join(api.portal.get().getPhysicalPath())
        kw['authorized_userid'] = api.user.get_current().getId()

        without_transaction = options.pop('without_transaction', False)

        celery = getCelery()
        # Here we cheat a little: since we will not start the task
        # up until the transaction is done,
        # we cannot give back to whoever called apply_async
        # its much beloved AsyncResult.
        # But we can actually pass the task a specific task_id
        # (although it's not very documented)
        # and an AsyncResult at this point is just that id, basically.
        task_id = uuid()

        # Construct a fake result
        if celery.conf.CELERY_ALWAYS_EAGER:
            result_ = EagerResult(task_id, None, states.PENDING, None)
        else:
            result_ = result.AsyncResult(task_id)

        # Note: one might be tempted to turn this into a datamanager.
        # This would result in two wrong things happening:
        # * A "commit within a commit" triggered by the function runner
        #   when CELERY_ALWAYS_EAGER is set,
        #   leading to the first invoked commit cleanup failing
        #   because the inner commit already cleaned up.
        # * An async task failing in eager mode would also rollback
        #   the whole transaction, which is not desiderable.
        #   Consider the case where the syncronous code constructs an object
        #   and the async task updates it, if we roll back everything
        #   then also the original content construction goes away
        #   (even if, in and by itself, worked)
        if without_transaction or celery.conf.CELERY_ALWAYS_EAGER:
            return self._apply_async(args, kw, result_, celery, task_id, options)
        else:
            queue_task_after_commit(args, kw, self, task_id, options)
            # Return the "fake" result ID
            return result_

    def _apply_async(self, args, kw, result_, celery, task_id, options):
        effective_result = super(AfterCommitTask, self).apply_async(
            args=args,
            kwargs=kw,
            task_id=task_id,
            **options
        )
        if celery.conf.CELERY_ALWAYS_EAGER:
            result_._state = effective_result._state
            result_._result = effective_result._result
            result_._traceback = effective_result._traceback
            celery.backend.store_result(
                task_id,
                effective_result._result,
                effective_result._state,
                traceback=result_.traceback,
                request=self.request
            )
            return result_
        return effective_result
